__all__ = ('create_event_loop', 'create_future', 'create_task', 'get_event_loop', 'run',)

from threading import current_thread, enumerate as list_threads

from ..utils import WeakItemDictionary
from .event_loop import EventThread
from .traps import Task, Future

THREAD_TO_EVENT_LOOP_REFERENCE = WeakItemDictionary()


def _try_detect_event_loop(local_thread):
    """
    Tries to detect the event loop, what should be used from the current thread.
    
    Returns
    -------
    event_loop : `None` or ``EventThread``
    """
    try:
        thread = THREAD_TO_EVENT_LOOP_REFERENCE[local_thread]
    except KeyError:
        pass
    else:
        if not thread.is_stopped():
            return thread
    
    for thread in list_threads():
        if not isinstance(thread, EventThread):
            continue
        
        if thread.daemon:
            continue
        
        if thread.is_stopped():
            continue
        
        THREAD_TO_EVENT_LOOP_REFERENCE[local_thread] = thread
        return thread


def _get_event_loop_is_current_thread():
    """
    Gets the local event loop if applicable and whether it is indeed the current thread or nah.
    
    Used by other top level functions.
    
    Returns
    -------
    event_loop : ``EventThread``
        The local event loop.
    is_current_thread : `bool`
        Whether the event loop is the current thread.
    
    Raises
    ------
    RuntimeError
        There are are no detectable event loops.
    """
    local_thread = current_thread()
    if isinstance(local_thread, EventThread):
        return local_thread, True
    
    event_loop = _try_detect_event_loop(local_thread)
    if (event_loop is not None):
        return event_loop, False
    
    raise RuntimeError(f'No running event loop on {local_thread!r}.')


def get_event_loop():
    """
    Gets the local event loop if applicable. If not raises `RuntimeError`.
    
    Returns
    -------
    event_loop : ``EventThread``
        The local event loop.
    
    Raises
    ------
    RuntimeError
        There are are no detectable event loops.
    """
    local_thread = current_thread()
    if isinstance(local_thread, EventThread):
        return local_thread
    
    event_loop = _try_detect_event_loop(local_thread)
    if (event_loop is not None):
        return event_loop
    
    raise RuntimeError(f'No running event loop on {local_thread!r}.')


def create_event_loop(**kwargs):
    """
    Creates a new event loop.
    
    Parameters
    ----------
    **kwargs : Keyword parameters
        Parameters to create the event loop with.
    
    Other parameters
    ----------------
    daemon : `bool`, Optional (Keyword only)
        Whether the event loop should be daemon. Defaults to `False`
    name : `None` or `str`, Optional (Keyword only)
        The event loop's name. Defaults ot `None`.
    start_later : `bool`, Optional (Keyword only)
        Whether the event loop should be started only later. Defaults to `True`
    keep_executor_count : `int`, Optional (Keyword only)
        The minimal amount of executors, what the event thread should keep alive. Defaults to `1`.

    Returns
    -------
    event_loop : ``EventThread``
    """
    return EventThread(**kwargs)


def create_task(coroutine):
    """
    Creates a task on the local event loop.
    
    Parameters
    ----------
    coroutine : `GeneratorType` or `CoroutineType`
        The coroutine to create task from.
    
    Returns
    -------
    task : ``Task``
    
    Raises
    ------
    RuntimeError
        There are are no detectable event loops.
    """
    loop, is_current_thread = _get_event_loop_is_current_thread()
    task = Task(coroutine, loop)
    if (not is_current_thread):
        loop.wake_up()
    
    return task


def create_future():
    """
    Creates s future bound to the local event loop.
    
    Returns
    -------
    task : ``Task``
    
    Raises
    ------
    RuntimeError
        There are are no detectable event loops.
    """
    loop, is_current_thread = _get_event_loop_is_current_thread()
    future = Future(loop)
    if (not is_current_thread):
        loop.wake_up()
    
    return future


def run(awaitable, timeout=None):
    """
    Runs the given awaitable on an event loop. This function cannot be called from an event loop itself.
    
    Parameters
    ----------
    awaitable : `awaitable`
        The awaitable to run.
    timeout : `None` or `float`, Optional
        Timeout after the awaitable should be cancelled. Defaults to `None`.
    
    Raises
    ------
    TypeError
        If `awaitable` is not `awaitable`.
    TimeoutError
         If `awaitable` did not finish before the given `timeout` is over.
    RuntimeError
        If called from itself.
    BaseException
        Any exception raised by `awaitable`.
    """
    local_thread = current_thread()
    if isinstance(local_thread, EventThread):
        raise RuntimeError(f'`{local_thread.__class__.__name__}.run` should not be called from itself.')
    
    event_loop = _try_detect_event_loop(local_thread)
    if (event_loop is None):
        stop_event_loop_after = True
        event_loop = EventThread(daemon=True, name='scarletio.run', start_later=False)
    else:
        stop_event_loop_after = False
    
    try:
        return event_loop.run(awaitable, timeout=timeout)
    finally:
        if stop_event_loop_after:
            event_loop.stop()



__all__ = ('RichAttributeErrorBaseType',)

from difflib import get_close_matches
from itertools import permutations as permutate

EXCEPTION_MESSAGE_CACHE = {}

def _iterate_alternative_attribute_names(attribute_name):
    """
    Tries to create alternative versions of the given `attribute_name`.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    attribute_name : `str`
        The attribute name to create alternative names of.
    
    Yields
    ------
    alternative_attribute_name : `str`
    """
    yield attribute_name
    
    split = attribute_name.split('_')
    split_count = len(split)
    if split_count > 1:
        for permutation in permutate(split, split_count):
            yield '_'.join(permutation)


def _create_rich_exception_message(instance, attribute_name):
    """
    Creates rich exception message for the given instance's given attribute's name.
    
    Parameters
    ----------
    instance : `object`
        The instance who has no attribute like `attribute_name`.
    attribute_name : `str`
        The evil variable name.
    """
    diversity = 0.8-(20-min(len(attribute_name), 20))*0.01
    directory = dir(instance)
    collected_matches = set()
    
    for alternative_attribute_name in _iterate_alternative_attribute_names(attribute_name):
        matches = get_close_matches(alternative_attribute_name, directory, n=10, cutoff=diversity)
        collected_matches.update(matches)
    
    sorted_matches = sorted(collected_matches)
    
    exception_message_parts = []
    
    exception_message_parts.append('`')
    exception_message_parts.append(type(instance).__name__)
    exception_message_parts.append('` has no attribute `')
    exception_message_parts.append(attribute_name)
    exception_message_parts.append('`')
    
    matches_count = len(sorted_matches)
    if matches_count == 0:
        pass
    elif matches_count == 1:
        exception_message_parts.append('; Did you mean: `')
        exception_message_parts.append(sorted_matches[0])
        exception_message_parts.append('`')
    
    else:
        exception_message_parts.append('; Did you mean any of: ')
        
        index = 0
        while True:
            exception_message_parts.append('`')
            exception_message_parts.append(sorted_matches[index])
            exception_message_parts.append('`')
            
            index += 1
            if index == matches_count:
                break
            
            exception_message_parts.append(', ')
            continue
    
    exception_message_parts.append('.')
    
    return ''.join(exception_message_parts)


class RichAttributeErrorBaseType:
    """
    Base type for generating rich attribute error messages.
    """
    __slots__ = ()
    
    def __getattr__(self, attribute_name):
        """Drops a rich attribute error."""
        if type(self).__getattr__ is RichAttributeErrorBaseType.__getattr__:
            key = (type(self).__module__, attribute_name)
            
            try:
                exception_message = EXCEPTION_MESSAGE_CACHE[key]
            except KeyError:
                exception_message = _create_rich_exception_message(self, attribute_name)
                EXCEPTION_MESSAGE_CACHE[key] = exception_message
        else:
            exception_message = _create_rich_exception_message(self, attribute_name)
        
        raise AttributeError(exception_message)



__all__ = ('skip_ready_cycle', 'sleep')

from threading import current_thread

from ...utils import to_coroutine, include

from .future import Future
from .handle_cancellers import _SleepHandleCanceller

EventThread = include('EventThread')


@to_coroutine
def skip_ready_cycle():
    """
    Skips a ready cycle.
    
    This function is a coroutine.
    """
    yield


def sleep(delay, loop=None):
    """
    Suspends the current task, allowing other tasks to run.
    
    Parameters
    ----------
    delay : `float`
        The time to block sleep in seconds.
    loop : `None` or ``EventThread``, Optional
        The event loop to which the returned `future` will be bound to. If not given, or given as `None`, then the
        current event loop will be used.
    
    Returns
    -------
    future : ``Future``
        Future, what can be awaited, to suspend a task.
    
    Raises
    ------
    RuntimeError
        The given or the local event loop is already stopped.
    """
    if loop is None:
        loop = current_thread()
        if not isinstance(loop, EventThread):
            raise RuntimeError(f'`sleep` called without passing `loop` parameter from a non {EventThread.__name__}: '
                f'{loop!r}.')
    
    future = Future(loop)
    if delay <= 0.:
        future.set_result(None)
        return future
    
    callback = object.__new__(_SleepHandleCanceller)
    handle = loop.call_later(delay, callback, future)
    if handle is None:
        raise RuntimeError(f'`sleep` was called with future with a stopped loop {loop!r}')
    
    future._callbacks.append(callback)
    callback._handle = handle
    return future




__all__ = ('IgnoreCaseMultiValueDictionary', )

from .docs import has_docs
from .multi_value_dictionary import MultiValueDictionary
from .ignore_case_string import IgnoreCaseString

@has_docs
class IgnoreCaseMultiValueDictionary(MultiValueDictionary):
    """
    ``MultiValueDictionary`` subclass, what can be used to hold http headers.
    
    It's keys ignore casing.
    """
    __slots__ = ()
    
    @has_docs
    def __init__(self, iterable=None):
        """
        Creates a new ``IgnoreCaseMultiValueDictionary`` instance.
        
        Parameters
        ----------
        iterable : `None` or `iterable`, Optional
            Iterable to update the created MultiValueDictionary initially.
            
            Can be given as one of the following:
                - ``MultiValueDictionary`` instance.
                - `dict` instance.
                - `iterable` of `key` - `value` pairs.
        """
        dict.__init__(self)
        
        if (iterable is None) or (not iterable):
            return
        
        getitem = dict.__getitem__
        setitem = dict.__setitem__
        
        if type(iterable) is type(self):
            for key, values in dict.items(iterable):
                setitem(self, key, values.copy())
        
        elif isinstance(iterable, MultiValueDictionary):
            for key, values in dict.items(iterable):
                setitem(self, IgnoreCaseString(key), values.copy())
        
        elif isinstance(iterable, dict):
            for key, value in iterable.items():
                key = IgnoreCaseString(key)
                setitem(self, key, [value])
        
        else:
            for key, value in iterable:
                key = IgnoreCaseString(key)
                try:
                    values = getitem(self, key)
                except KeyError:
                    setitem(self, key, [value])
                else:
                    values.append(value)
    
    @has_docs
    def __getitem__(self, key):
        """
        Returns the MultiValueDictionary's `value` for the given `key`. If the `key` has more values, then returns the 0th of
        them.
        """
        key = IgnoreCaseString(key)
        return dict.__getitem__(self, key)[0]
    
    @has_docs
    def __setitem__(self, key, value):
        """Adds the given `key` - `value` pair to the MultiValueDictionary."""
        key = IgnoreCaseString(key)
        MultiValueDictionary.__setitem__(self, key, value)
    
    @has_docs
    def __delitem__(self, key):
        """
        Removes the `value` for the given `key` from the MultiValueDictionary. If the `key` has more values, then removes only
        the 0th of them.
        """
        key = IgnoreCaseString(key)
        MultiValueDictionary.__delitem__(self, key)
    
    @has_docs
    def extend(self, mapping):
        """
        Extends the MultiValueDictionary titled with the given `mapping`'s items.
        
        Parameters
        ----------
        mapping : `Any`
            Any mapping type, what has `.items` attribute.
        """
        getitem = dict.__getitem__
        setitem = dict.__setitem__
        for key, value in mapping.items():
            key = IgnoreCaseString(key)
            try:
                values = getitem(self, key)
            except KeyError:
                setitem(self, key, [value])
            else:
                if value not in values:
                    values.append(value)
    
    @has_docs
    def get_all(self, key, default=None):
        """
        Returns all the values matching the given `key`.
        
        Parameters
        ----------
        key : `Any`
            The `key` to match.
        default : `Any`, Optional
            Default value to return if `key` is not present in the MultiValueDictionary. Defaults to `None`.
        
        Returns
        -------
        values : `default or `list` of `Any`
            The values for the given `key` if present.
        """
        key = IgnoreCaseString(key)
        return MultiValueDictionary.get_all(self, key, default)
    
    @has_docs
    def get_one(self, key, default=None):
        """
        Returns the 0th value matching the given `key`.
        
        Parameters
        ----------
        key : `Any`
            The key to match.
        default : `Any`, Optional
            Default value to return if `key` is not present in the MultiValueDictionary. Defaults to `None`.
        
        Returns
        -------
        value : `default` or `Any`
            The value for the given key if present.
        """
        key = IgnoreCaseString(key)
        return MultiValueDictionary.get_one(self, key, default)
    
    get = get_one
    
    @has_docs
    def setdefault(self, key, default=None):
        """
        Returns the value for the given `key`.
        
        If the `key` is not present in the MultiValueDictionary, then set's the given `default` value as it.
        
        Parameters
        ----------
        key : `Any`
            The key to match.
        default : `Any`, Optional
            Default value to set and return if `key` is not present in the MultiValueDictionary.
        
        Returns
        -------
        value : `default` or `Any`
            The first value for which `key` matched, or `default` if none.
        """
        key = IgnoreCaseString(key)
        return MultiValueDictionary.setdefault(self, key, default)
    
    @has_docs
    def pop_all(self, key, default=...):
        """
        Removes all the values from the MultiValueDictionary which the given `key` matched.
        
        Parameters
        ----------
        key : `Any`
            The key to match.
        default : `Any`, Optional
            Default value to return if `key` is not present in the MultiValueDictionary.
        
        Returns
        -------
        values : `default` or `list` of `Any`
            The matched values. If `key` is not present, but `default` value is given, then returns that.
        
        Raises
        ------
        KeyError
            if `key` is not present in the MultiValueDictionary and `default` value is not given either.
        """
        key = IgnoreCaseString(key)
        return MultiValueDictionary.pop_all(self, key, default)
    
    @has_docs
    def pop_one(self, key, default=...):
        """
        Removes the first value from the MultiValueDictionary, which matches the given `key`.
        
        Parameters
        ----------
        key : `Any`
            The key to match.
        default : `Any`, Optional
            Default value to return if `key` is not present in the MultiValueDictionary.
        
        Returns
        -------
        value : `default` or `list` of `Any`
            The 0th matched value. If `key` is not present, but `default` value is given, then returns that.
        
        Raises
        ------
        KeyError
            if `key` is not present in the MultiValueDictionary and `default` value is not given either.
        """
        key = IgnoreCaseString(key)
        return MultiValueDictionary.pop_one(self, key, default)
    
    pop = pop_one



__all__ = ('ignore_frame', 'render_exception_into', 'render_frames_into', 'should_ignore_frame',)

import reprlib, linecache
from types import GeneratorType, CoroutineType, MethodType

IGNORED_FRAME_LINES = {}

def try_get_raw_exception_representation(exception):
    """
    Tries to get raw exception representation.
    
    Parameters
    ----------
    exception : ``BaseException``
        The respective exception instance.
    
    Returns
    -------
    raw_exception_representation : `str`
    """
    raw_exception_representation_parts = [
        '> repr(exception) raised, trying to get raw representation.\n'
    ]
    
    exception_name = getattr(type(exception), '__name__')
    if type(exception_name) is str:
        pass
    elif isinstance(exception_name, str):
        try:
            exception_name = str(exception_name)
        except:
            exception_name = '<Exception>'
    else:
        exception_name = '<Exception>'
    
    raw_exception_representation_parts.append(exception_name)
    raw_exception_representation_parts.append('(')
    
    try:
        args = getattr(exception, 'args', None)
    except:
        pass
    else:
        if (args is not None) and (type(args) is tuple):
            length = len(args)
            if length:
                index = 0
                while True:
                    element = args[index]
                    
                    try:
                        element_representation = repr(element)
                    except:
                        element_representation = f'<parameter_{index}>'
                    else:
                        if type(element_representation) is not str:
                            try:
                                element_representation = str(element_representation)
                            except:
                                element_representation = f'<parameter_{index}>'
                    
                    raw_exception_representation_parts.append(element_representation)
                    
                    index += 1
                    if index == length:
                        break
                    
                    raw_exception_representation_parts.append(', ')
                    continue
        
    
    raw_exception_representation_parts.append(')')
    return ''.join(raw_exception_representation_parts)


def get_exception_representation(exception):
    """
    Gets the exception's representation.
    
    Parameters
    ----------
    exception : ``BaseException``
        The respective exception instance.
    
    Returns
    -------
    exception_representation : `str`
    """
    try:
        exception_representation = repr(exception)
    except:
        exception_representation = try_get_raw_exception_representation(exception)
    
    return exception_representation


def ignore_frame(file, name, line):
    """
    When rendering an exception traceback, specified frames can be added to being stopped from rendering.
    
    Parameters
    ----------
    file : `str`
        The name of the respective file.
    name : `str`
        The name of the respective function.
    line : `str`
        The respective line's stripped content.
    """
    try:
        files = IGNORED_FRAME_LINES[file]
    except KeyError:
        files = {}
        IGNORED_FRAME_LINES[file] = files

    try:
        names = files[name]
    except KeyError:
        names = set()
        files[name] = names

    names.add(line)


def should_ignore_frame(file, name, line):
    """
    Returns whether the given frame should be ignored from rending.
    
    Parameters
    ----------
    file : `str`
        The frame's respective file's name.
    name : `str`
        The frame's respective function's name.
    line : `str`
        The frame's respective stripped line.
    
    Returns
    -------
    should_ignore : `bool`
    """
    try:
        files = IGNORED_FRAME_LINES[file]
    except KeyError:
        return False

    try:
        names = files[name]
    except KeyError:
        return False
    
    return (line in names)


def format_callback(func, args=None, kwargs=None):
    """
    Formats the given callback to a more user friendly representation.
    
    Parameters
    ----------
    func : `callable`
        The callback to format.
    args : `None` or `iterable` of `Any`, Optional
        Additional parameters to call the `func` with.
    kwargs : `None` or `dict` of (`str`, `Any`) items, Optional
        Additional keyword parameters to call the `func` with.
    
    Returns
    -------
    result : `str`
        The formatted callback.
    """
    result = []
    # un-warp the wrappers
    while True:
        if not (None is args is kwargs):
            sub_result = ['(']
            if (args is not None) and args:
                for arg in args:
                    sub_result.append(reprlib.repr(arg))
                    sub_result.append(', ')
            
            if (kwargs is not None) and kwargs:
                for key, arg in kwargs.items():
                    sub_result.append(str(key)) # never trust
                    sub_result.append('=')
                    sub_result.append(reprlib.repr(arg))
                    sub_result.append(', ')
            
            if len(sub_result) > 1:
                del sub_result[-1]
            
            sub_result.append(')')
            result.append(''.join(sub_result))
        
        try:
            wrapped = func.func
        except AttributeError:
            if (type(func) is MethodType) and hasattr(type(func.__self__), 'get_coroutine'):
                coroutine = func.__self__.get_coroutine()
                coroutine_repr = getattr(coroutine, '__qualname__', None)
                if coroutine_repr is None:
                    coroutine_repr = getattr(coroutine, '__name__', None)
                    if coroutine_repr is None:
                        coroutine_repr = repr(coroutine)
                func_repr = f'<Bound method {func.__func__.__name__} of Task {coroutine_repr}>'
            else:
                func_repr = getattr(func, '__qualname__', None)
                if func_repr is None:
                    func_repr = getattr(func, '__name__', None)
                    if func_repr is None:
                        func_repr = repr(func)
            
            result.insert(0, func_repr)
            break
        
        args = getattr(func, 'args', None)
        kwargs = getattr(func, 'kwargs', None)
        func = wrapped
    
    return ''.join(result)


def format_coroutine(coroutine):
    """
    Formats the given coroutine to a more user friendly representation.
    
    Parameters
    ----------
    coroutine : `CoroutineType` or `GeneratorType` (or any compatible cyi or builtin)
        The coroutine to get representation of.
    
    Returns
    -------
    result : `str`
        The formatted coroutine.
    """
    if not (hasattr(coroutine, 'cr_code') or hasattr(coroutine, 'gi_code')):
        # Cython or builtin
        name = getattr(coroutine, '__qualname__', None)
        if name is None:
            name = getattr(coroutine, '__name__', None)
            if name is None: # builtins might reach this part
                name = coroutine.__class__.__name__
        
        if type(coroutine) is GeneratorType:
            running = coroutine.gi_running
        elif type(coroutine) is CoroutineType:
            running = coroutine.cr_running
        else:
            running = False
        
        if running:
            state = 'running'
        else:
            state = 'done'
        
        return f'{name}() {state}'
    
    name = format_callback(coroutine)
    
    if type(coroutine) is GeneratorType:
        code = coroutine.gi_code
        frame = coroutine.gi_frame
    else:
        code = coroutine.cr_code
        frame = coroutine.cr_frame
    
    file_name = code.co_filename
    
    if frame is None:
        line_number = code.co_firstlineno
        state = 'done'
    else:
        line_number = frame.f_lineno
        state = 'running'
    
    return f'{name} {state} defined at {file_name}:{line_number}'


def render_frames_into(frames, extend=None):
    """
    Renders the given frames into a list of strings.
    
    Parameters
    ----------
    frames : `list` of (`frame` or ``ExceptionFrameProxy``)
        The frames to render.
    extend : `None` or `list` of `str`
        Whether the frames should be rendered into an already existing list.
    
    Returns
    -------
    extend : `list` of `str`
        The rendered frames as a `list` of it's string parts.
    """
    if extend is None:
        extend = []
    
    checked = set()
    
    last_file_name = ''
    last_line_number = ''
    last_name = ''
    count = 0
    
    for frame in frames:
        line_number = frame.f_lineno
        code = frame.f_code
        file_name = code.co_filename
        name = code.co_name
        
        if last_file_name == file_name and last_line_number == line_number and last_name == name:
            count += 1
            if count > 2:
                continue
        else:
            if count > 3:
                count -= 3
                extend.append('  [Previous line repeated ')
                extend.append(str(count))
                extend.append(' more times]\n')
            count = 0
        
        if file_name not in checked:
            checked.add(file_name)
            linecache.checkcache(file_name)
        
        line = linecache.getline(file_name, line_number, None)
        line = line.strip()
        
        if should_ignore_frame(file_name, name, line):
            continue
        
        last_file_name = file_name
        last_line_number = line_number
        last_name = code.co_name
        
        extend.append('  File \"')
        extend.append(file_name)
        extend.append('\", line ')
        extend.append(str(line_number))
        extend.append(', in ')
        extend.append(name)
        if (line is not None) and line:
            extend.append('\n    ')
            extend.append(line)
        extend.append('\n')
        
    if count > 3:
        count -= 3
        extend.append('  [Previous line repeated ')
        extend.append(str(count))
        extend.append(' more times]\n')
    
    return extend


class ExceptionFrameProxy:
    """
    Wraps a `traceback` object to be `frame` compatible.
    
    Attributes
    ----------
    tb : `traceback`
        The wrapped traceback frame.
    """
    __slots__ = ('tb',)
    
    def __init__(self, tb):
        """
        Creates a new ``ExceptionFrameProxy`` instance with the given traceback.
        
        tb : `traceback`
            The traceback to wrap.
        """
        self.tb = tb
    
    @property
    def f_builtins(self):
        """
        Returns the traceback's frame's builtins.
        
        Returns
        -------
        f_builtins : `dict` of (`str`, `Any`) items
        """
        return self.tb.tb_frame.f_builtins
    
    @property
    def f_code(self):
        """
        Returns the traceback's frame's code.
        
        Returns
        -------
        f_code : `code`
        """
        return self.tb.tb_frame.f_code
    
    @property
    def f_globals(self):
        """
        Returns the traceback's frame's globals.
        
        Returns
        -------
        f_globals : `dict` of (`str`, `Any`)
        """
        return self.tb.tb_frame.f_globals
    
    @property
    def f_lasti(self):
        """
        Returns the traceback's frame's last attempted instruction index in the bytecode.
        
        Returns
        -------
        f_lasti : `int`
        """
        return self.tb.tb_frame.f_lasti
    
    @property
    def f_lineno(self):
        """
        Returns the traceback's frame's current line number in Python source code.
        
        Returns
        -------
        f_lineno : `int`
        """
        return self.tb.tb_lineno
    
    @property
    def f_locals(self):
        """
        Returns the local variables, what the traceback's frame can see.
        
        Returns
        -------
        f_locals : `dict` of (`str`, `Any`)
        """
        return self.tb.tb_frame.f_locals
    
    @property
    def f_trace(self):
        """
        Tracing function for the traceback's frame.
        
        Returns
        -------
        f_trace : `Any`
            Defaults to `None`.
        """
        return self.tb.tb_frame.f_trace

def _get_exception_frames(exception):
    """
    Gets the frames of the given exception.
    
    Parameters
    ----------
    exception : `BaseException` instance
        The exception to trace back.
    
    Returns
    -------
    frames : `list` of ``ExceptionFrameProxy``
        A list of `frame` compatible exception frames.
    """
    frames = []
    tb = exception.__traceback__
    
    while True:
        if tb is None:
            break
        frame = ExceptionFrameProxy(tb)
        frames.append(frame)
        tb = tb.tb_next
    
    return frames

def render_exception_into(exception, extend=None):
    """
    Renders the given exception's frames into a list of strings.
    
    Parameters
    ----------
    exception : `BaseException` instance
        The exception to render.
    extend : `None` or `list` of `str`
        Whether the frames should be rendered into an already existing list.
    
    Returns
    -------
    extend : `list` of `str`
        The rendered frames as a `list` of it's string parts.
    """
    if extend is None:
        extend = []
    
    exceptions = []
    reason_type = 0
    while True:
        exceptions.append((exception, reason_type))
        cause_exception = exception.__cause__
        if (cause_exception is not None):
            exception = cause_exception
            reason_type = 1
            continue
        
        context_exception = exception.__context__
        if (context_exception is not None):
            exception = context_exception
            reason_type = 2
            continue
        
        # no other cases
        break
    
    for exception, reason_type in reversed(exceptions):
        frames = _get_exception_frames(exception)
        extend.append('Traceback (most recent call last):\n')
        extend = render_frames_into(frames, extend=extend)
        extend.append(get_exception_representation(exception))
        extend.append('\n')
        
        if reason_type == 0:
            break
        
        if reason_type == 1:
            extend.append('\nThe above exception was the direct cause of the following exception:\n\n')
            continue
        
        if reason_type == 2:
            extend.append('\nDuring handling of the above exception, another exception occurred:\n\n')
            continue
        
        # no more cases
        continue
    
    return extend



__all__ = ()

from .headers import CONNECTION, TRANSFER_ENCODING, CONTENT_ENCODING

class RawMessage:
    """
    Base class of ``RawResponseMessage`` and ``RawRequestMessage``.
    
    Attributes
    ----------
    _upgraded : `int`
        Whether the connection is upgraded.
        
        Is only set when the ``.upgraded`` property is accessed as `0` or `1`. Till is set as `2`.
    headers : ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
        The headers of the http message.
    """
    __slots__ = ('_upgraded', 'headers', )
    
    @property
    def upgraded(self):
        """
        A get-set descriptor to access whether the message is upgraded. On first access the upgrade state is detected
        from the headers.
        """
        upgraded = self._upgraded
        if upgraded == 2:
            try:
                connection = self.headers[CONNECTION]
            except KeyError:
                upgraded = 0
            else:
                upgraded = (connection.lower() == 'upgrade')
            
            self._upgraded = upgraded
        
        return upgraded
    
    
    @upgraded.setter
    def upgraded(self, value):
        self._upgraded = value
    
    
    @property
    def chunked(self):
        """
        Returns whether the respective http message' body is chunked.
        
        Returns
        -------
        chunked : `bool`
        """
        try:
            transfer_encoding = self.headers[TRANSFER_ENCODING]
        except KeyError:
            return False
        
        return ('chunked' in transfer_encoding.lower())
    
    
    @property
    def encoding(self):
        """
        Returns the body encoding set in the http message's header.
        
        Returns
        -------
        encoding : `None` or `str`
            If no encoding is set, defaults to `None`.
        """
        try:
            encoding = self.headers[CONTENT_ENCODING]
        except KeyError:
            encoding = None
        else:
            encoding = encoding.lower()
        
        return encoding


class RawResponseMessage(RawMessage):
    """
    Represents a raw http response message.
    
    Attributes
    ----------
    _upgraded : `int`
        Whether the connection is upgraded.
        
        Is only set when the ``.upgraded`` property is accessed as `0` or `1`. Till is set as `2`.
    headers : ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
        The headers of the http message.
    version : ``HttpVersion``
        The http version of the response.
    status : `int`
        The response's status.
    reason : `bytes`
        Reason included with the response. Might be empty.
    """
    __slots__ = ('version', 'status', 'reason',)
    
    def __init__(self, version, status, reason, headers):
        """
        Creates a new ``RawResponseMessage`` instance with the given parameters.
        
        Parameters
        ----------
        version : ``HttpVersion``
            The http version of the response.
        status : `int`
            The response's status.
        reason : `bytes`
            Reason included with the response. Might be empty.
        headers : ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
            The headers of the http message.
        """
        self.version = version
        self.status = status
        self.reason = reason
        self.headers = headers
        self._upgraded = 2


class RawRequestMessage(RawMessage):
    """
    Represents a raw http request message.
    
    Attributes
    ----------
    _upgraded : `int`
        Whether the connection is upgraded.
        
        Is only set when the ``.upgraded`` property is accessed as `0` or `1`. Till is set as `2`.
    headers : ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
        The headers of the http message.
    version : ``HttpVersion``
        The http version of the response.
    method : `int`
        The request's method.
    path : `str`
        The requested path.
    """
    __slots__ = ('version', 'method', 'path',)
    
    def __init__(self, version, method, path, headers):
        """
        Creates a new ``RawRequestMessage`` instance with the given parameters.
        
        Parameters
        ----------
        version : ``HttpVersion``
            The http version of the response.
        method : `int`
            The request's method.
        path : `str`
            The requested path.
        headers : ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
            The headers of the http message.
        """
        self.version = version
        self.method = method
        self.path = path
        self.headers = headers
        self._upgraded = 2
